@@ -1,4 +1,5 @@
 #include <caml/alloc.h>
+#include <caml/bigarray.h>
 #include <caml/callback.h>
 #include <caml/custom.h>
 #include <caml/fail.h>
@@ -358,6 +359,79 @@ CAMLprim value ocaml_opus_decoder_decode_float_byte(value *argv, int argn) {
                                          argv[4], argv[5]);
 }
 
+CAMLprim value ocaml_opus_decoder_decode_float_ba(value _dec, value _os,
+                                                  value buf, value _ofs,
+                                                  value _len, value _fec) {
+  CAMLparam3(_dec, _os, buf);
+  CAMLlocal1(chan);
+  ogg_stream_state *os = Stream_state_val(_os);
+  ogg_packet op;
+  OpusDecoder *dec = Dec_val(_dec);
+  int decode_fec = Int_val(_fec);
+
+  int ofs = Int_val(_ofs);
+  int len = Int_val(_len);
+  int total_samples = 0;
+  int ret;
+
+  int chans = Wosize_val(buf);
+  float *pcm = malloc(chans * len * sizeof(float));
+  if (pcm == NULL)
+    caml_raise_out_of_memory();
+
+  int i, c;
+
+  while (total_samples < len) {
+    ret = ogg_stream_packetout(os, &op);
+    /* returned values are:
+     * 1: ok
+     * 0: not enough data. in this case
+     *    we return the number of samples
+     *    decoded if > 0 and raise
+     *    Ogg_not_enough_data otherwise
+     * -1: out of sync */
+    if (ret == -1)
+      caml_raise_constant(*caml_named_value("ogg_exn_out_of_sync"));
+
+    if (ret == 0) {
+      free(pcm);
+      if (total_samples > 0) {
+        CAMLreturn(Val_int(total_samples));
+      } else {
+        caml_raise_constant(*caml_named_value("ogg_exn_not_enough_data"));
+      }
+    }
+
+    if (chans != opus_packet_get_nb_channels(op.packet))
+      caml_invalid_argument("Wrong number of channels.");
+
+    caml_release_runtime_system();
+    ret = opus_decode_float(dec, op.packet, op.bytes, pcm, len, decode_fec);
+    caml_acquire_runtime_system();
+
+    if (ret < 0) {
+      free(pcm);
+      check(ret);
+    }
+    for (c = 0; c < chans; c++) {
+      chan = Field(buf, c);
+      for (i = 0; i < ret; i++)
+        ((float *)Caml_ba_data_val(chan))[ofs + total_samples + i] =
+            pcm[i * chans + c];
+    }
+    total_samples += ret;
+    len -= ret;
+  }
+
+  free(pcm);
+  CAMLreturn(Val_int(total_samples));
+}
+
+CAMLprim value ocaml_opus_decoder_decode_float_ba_byte(value *argv, int argn) {
+  return ocaml_opus_decoder_decode_float_ba(argv[0], argv[1], argv[2], argv[3],
+                                            argv[4], argv[5]);
+}
+
 /***** Encoder *****/
 
 typedef struct encoder_t {
@@ -404,7 +478,7 @@ static value value_of_application(opus_int32 a) {
   }
 }
 
-static char header_packet[19] = {
+static unsigned char header_packet[19] = {
     /* Identifier. */
     'O', 'p', 'u', 's', 'H', 'e', 'a', 'd',
     /* version, channels count, pre-skip (16 bits, unsigned,
@@ -419,11 +493,7 @@ static char header_packet[19] = {
 static void pack_header(ogg_packet *op, opus_int32 sr, int channels,
                         opus_int16 pre_skip, opus_int16 gain) {
   op->bytes = sizeof(header_packet);
-  op->packet = malloc(op->bytes);
-  if (op->packet == NULL)
-    caml_raise_out_of_memory();
-
-  memcpy(op->packet, header_packet, op->bytes);
+  op->packet = header_packet;
 
   /* Now fill data. */
   op->packet[9] = channels;
@@ -520,6 +590,8 @@ CAMLprim value ocaml_opus_encoder_create(value _skip, value _comments,
   Store_field(ans, 1, value_of_packet(&header));
   Store_field(ans, 2, value_of_packet(&comments));
 
+  free(comments.packet);
+
   CAMLreturn(ans);
 }
 
@@ -662,8 +734,8 @@ CAMLprim value ocaml_opus_encoder_ctl(value ctl, value _enc) {
   caml_failwith("Unknown opus error");
 }
 
-CAMLprim value ocaml_opus_encode_float(value _frame_size, value _enc, value buf,
-                                       value _off, value _len, value _os) {
+CAMLprim value ocaml_opus_encode_float(value _frame_size, value _enc, value _os,
+                                       value buf, value _off, value _len) {
   CAMLparam3(_enc, buf, _os);
   encoder_t *handler = Enc_val(_enc);
   OpusEncoder *enc = handler->encoder;
@@ -683,7 +755,7 @@ CAMLprim value ocaml_opus_encode_float(value _frame_size, value _enc, value buf,
   if (data == NULL)
     caml_raise_out_of_memory();
   float *pcm = malloc(chans * frame_size * sizeof(float));
-  if (data == NULL)
+  if (pcm == NULL)
     caml_raise_out_of_memory();
   int i, j, c;
   int ret;
@@ -701,8 +773,8 @@ CAMLprim value ocaml_opus_encode_float(value _frame_size, value _enc, value buf,
     if (ret < 0) {
       free(pcm);
       free(data);
+      check(ret);
     }
-    check(ret);
 
     /* From the documentation: If the return value is 1 byte,
      * then the packet does not need to be transmitted (DTX). */
@@ -718,9 +790,11 @@ CAMLprim value ocaml_opus_encode_float(value _frame_size, value _enc, value buf,
     op.packetno = handler->packetno;
     op.granulepos = handler->granulepos;
 
-    if (ogg_stream_packetin(os, &op) != 0)
+    if (ogg_stream_packetin(os, &op) != 0) {
+      free(pcm);
+      free(data);
       caml_raise_constant(*caml_named_value("ogg_exn_internal_error"));
-    ;
+    }
   }
   free(pcm);
   free(data);
@@ -733,6 +807,86 @@ CAMLprim value ocaml_opus_encode_float_byte(value *argv, int argn) {
                                  argv[5]);
 }
 
+CAMLprim value ocaml_opus_encode_float_ba(value _frame_size, value _enc,
+                                          value _os, value buf, value _ofs,
+                                          value _len) {
+  CAMLparam3(_enc, buf, _os);
+  encoder_t *handler = Enc_val(_enc);
+  OpusEncoder *enc = handler->encoder;
+  ogg_stream_state *os = Stream_state_val(_os);
+  ogg_packet op;
+  int len = Int_val(_len);
+  int ofs = Int_val(_ofs);
+  int chans = Wosize_val(buf);
+  int frame_size = Int_val(_frame_size);
+
+  if (chans == 0)
+    CAMLreturn(Val_int(0));
+
+  if (Caml_ba_array_val(Field(buf, 0))->dim[0] < ofs + len)
+    caml_failwith("Invalid length or offset!");
+
+  if (len < frame_size)
+    caml_raise_constant(*caml_named_value("opus_exn_buffer_too_small"));
+
+  /* This is the recommended value */
+  int max_data_bytes = 4000;
+  unsigned char *data = malloc(max_data_bytes);
+  if (data == NULL)
+    caml_raise_out_of_memory();
+  float *pcm = malloc(chans * frame_size * sizeof(float));
+  if (pcm == NULL)
+    caml_raise_out_of_memory();
+  int i, j, c;
+  int ret;
+  int loops = len / frame_size;
+  for (i = 0; i < loops; i++) {
+    for (j = 0; j < frame_size; j++)
+      for (c = 0; c < chans; c++)
+        pcm[chans * j + c] = ((float *)Caml_ba_data_val(
+            Field(buf, c)))[j + i * frame_size + ofs];
+
+    caml_release_runtime_system();
+    ret = opus_encode_float(enc, pcm, frame_size, data, max_data_bytes);
+    caml_acquire_runtime_system();
+
+    if (ret < 0) {
+      free(pcm);
+      free(data);
+      check(ret);
+    }
+
+    /* From the documentation: If the return value is 1 byte,
+     * then the packet does not need to be transmitted (DTX). */
+    if (ret < 2)
+      continue;
+
+    handler->granulepos += frame_size * handler->samplerate_ratio;
+    handler->packetno++;
+
+    op.bytes = ret;
+    op.packet = data;
+    op.b_o_s = op.e_o_s = 0;
+    op.packetno = handler->packetno;
+    op.granulepos = handler->granulepos;
+
+    if (ogg_stream_packetin(os, &op) != 0) {
+      free(pcm);
+      free(data);
+      caml_raise_constant(*caml_named_value("ogg_exn_internal_error"));
+    }
+  }
+  free(pcm);
+  free(data);
+
+  CAMLreturn(Val_int(loops * frame_size));
+}
+
+CAMLprim value ocaml_opus_encode_float_ba_byte(value *argv, int argn) {
+  return ocaml_opus_encode_float_ba(argv[0], argv[1], argv[2], argv[3], argv[4],
+                                    argv[5]);
+}
+
 CAMLprim value ocaml_opus_encode_eos(value _os, value _enc) {
   CAMLparam2(_os, _enc);
   ogg_stream_state *os = Stream_state_val(_os);
